from powerml import PowerML
import logging

logger = logging.getLogger(__name__)


class WriteEmailModel:
    def __init__(self, config={}, max_output_tokens=512, temperature=0.7):
        self.max_output_tokens = max_output_tokens
        self.model = PowerML(config)
        self.model_name = "stensul/email-copy"
        self.temperature = temperature

    def predict(self, input):
        output = self.model.predict(
            input,
            max_tokens=self.max_output_tokens,
            temperature=self.temperature,
            model=self.model_name
        )
        return output



from powerml import Filter


class ForecastFilter(Filter):
    '''
    This is a class that can be used to filter noise from data for ForecastSequenceModels.
    '''

    def __init__(self):
        # TODO: figure out a prompt that works
        super().__init__('text-davinci-002', 'daily revenue forecast for the podcast')




from powerml import PowerML
import logging

logger = logging.getLogger(__name__)


class ForecastSequenceModel:
    def __init__(self,
                 config={},
                 max_output_tokens=256,
                 ):
        self.model = PowerML(config)
        self.model_name = "inference-io/forecast-wondery"
        self.max_output_tokens = max_output_tokens

    def predict(self, prompt):
        output = self.model.predict(
            prompt, max_tokens=self.max_output_tokens, temperature=0.7, model=self.model_name)
        return self.__post_process(output)

    def __post_process(self, series):
        series = series.strip()
        series = series.removesuffix(",")
        return [val.strip() for val in series.split(",")]



from powerml import Generator
from powerml import CreateTopicsModel
import nltk
from nltk.stem import WordNetLemmatizer

from math import ceil

unblocked_human_unprompted_1gram_topics = [
    "SourceMarks",
    "VSCode",
    "Web",
    "Dashboard",
    "Kotlin",
    "Styles",
    "Services",
    "Video",
    "Messages",
    "Threads",
    "Mentions",
    "Git",
    "GitHub",
    "Auth",
    "Hub",
    "Slack",
    "Webhooks",
    "Gradle",
    "Intercom",
    "Adminconsole",
    "Ingestion",
    "Api",
    "Onboarding",
    "Discussions",
    "Insights",
    "Email",
    "Teams",
    "Notifications",
    "Pusher",
    "Logging",
    "Security",
    "Commands",
    "Controllers",
    "Stores",
    "Datastores",
    "Test",
    "Typescript",
    "React",
    "Recommendations",
    "Logs",
    "Compression",
    "Compute",
    "Webpack",
    "Webextension"]


class TopicGenerator(Generator):
    '''
    This is a class that can be used to generate more messages for CreateTopicsModels.
    '''

    def __init__(self, gold_topics=unblocked_human_unprompted_1gram_topics):
        nltk.download('wordnet')
        self.lemmatizer = WordNetLemmatizer()
        super().__init__([self.lemmatizer.lemmatize(topic.lower())
                          for topic in gold_topics])

    def coverage_generator(self, messages, return_metrics=True):
        model = CreateTopicsModel()
        model.fit(messages, 'one-word system components')
        generated_topics = [self.lemmatizer.lemmatize(
            topic.lower()) for topic in model.predict()]
        metrics = self.compute_coverage(generated_topics)
        coverage = metrics['Coverage']
        rare_topics = metrics['Rare Types']
        generated_messages = []
        if rare_topics:
            # generate at least 1 example per rare_topic, proportional to the amount of coverage of the messages
            num_generate = ceil(
                (1 - coverage) * len(messages) / len(rare_topics))
            print(f'num generate: {num_generate}')
            for topic in rare_topics:
                modifier = f'include the topic \'{topic}\''
                generated_messages.extend(self._generate_modified(
                    messages, modifier, num_generate))
        if return_metrics:
            return generated_messages, metrics
        return generated_messages



from powerml import Generator
from powerml import ForecastSequenceModel

class ForecastGenerator(Generator):
    '''
    This is a class that can be used to generate more data for ForecastSequenceModels.
    '''

    def __init__(self, gold_types=['News']):
        super().__init__(gold_types)

    def coverage_generator(self, data, return_metrics=True):
        return super().coverage_generator(data, return_metrics, ForecastSequenceModel())



from powerml import PowerML
import logging

logger = logging.getLogger(__name__)


class QuestionAnswerModel:
    def __init__(self,
                 config={},
                 max_output_tokens=256,
                 ):
        self.model = PowerML(config)
        self.model_name = "quizlet/questions"
        self.max_output_tokens = max_output_tokens

    def fit(self,
            lesson: str,
            answer: str,
            examples: list[str]):
        self.lesson = lesson
        self.question_count = str(len(examples) + 1)
        self.answer = answer
        self.examples = examples

    def predict(self):
        examples_string = "\n".join(
            [f'{i + 1}. ' + example for i, example in enumerate(self.examples)])
        prompt = {
            "{{lesson}}": self.lesson,
            "{{question_count}}": self.question_count,
            "{{answer}}": self.answer,
            "{{examples}}": examples_string,
        }
        output = self.model.predict(
            prompt, max_tokens=self.max_output_tokens, temperature=0.7, model=self.model_name)
        return self.__post_process(output)

    def __post_process(self, output):
        return output.strip()



from powerml import Generator
from powerml import AutocompleteSQLModel
from powerml.utils.generator import create_schema, tables_to_schema
from math import ceil

class SQLGenerator(Generator):
    '''
    This is a class that can be used to generate more queries for AutocompleteSQLModels.
    '''

    def __init__(self, schema):
        super().__init__(schema, 'code-davinci-002', max_output_tokens=1250)

    def coverage_generator(self, queries, return_metrics=True):
        model = AutocompleteSQLModel()
        model.fit(tables_to_schema(self.gold_types), queries)
        model_predictions = []
        for query in queries:
            query = query.split()[0]
            model_predictions.append(query + ' ' + model.predict(query))
        generated_schema = create_schema(model_predictions)
        metrics = self.compute_coverage(generated_schema)
        coverage = metrics['Coverage']
        rare_tables = metrics['Rare Types']
        generated_queries = []
        if rare_tables:
            # generate at least 1 example per rare_table, proportional to the amount of coverage of the queries
            num_generate = ceil((1 - coverage) * len(queries) / len(rare_tables))
            print(f'num generate: {num_generate}')
            for table in rare_tables:
                modifier = f'include \'{table}\''
                generated_queries.extend(self._generate_modified(queries, modifier, num_generate))
        if return_metrics:
            return generated_queries, metrics
        return generated_queries


from powerml import Filter

class EmailFilter(Filter):
    '''
    This is a class that can be used to filter noise from data for WriteEmailModels.
    '''

    def __init__(self):
        super().__init__('text-davinci-002', 'marketing email for the company')


from powerml import Generator
from powerml import QuestionAnswerModel
from powerml.utils.generator import get_types
from math import ceil

class QuestionGenerator(Generator):
    '''
    This is a class that can be used to generate more data for QuestionAnswerModels.
    '''

    def __init__(self, gold_types=['True-or-False']):
        super().__init__(gold_types)

    def coverage_generator(self, data, return_metrics=True):
        model = QuestionAnswerModel()
        model_predictions = []
        for datum in data:
            model.fit(datum[0], datum[2], [datum[1]])
            model_predictions.append(model.predict())
        print(f'model predictions: {model_predictions}')
        generated_types = get_types(model_predictions, self.gold_types)
        metrics = self.compute_coverage(generated_types)
        coverage = metrics['Coverage']
        rare_types = metrics['Rare Types']
        generated_data = []
        if rare_types:
            # generate at least 1 example per rare_type, proportional to the amount of coverage of the data
            num_generate = ceil((1 - coverage) * len(data) / len(rare_types))
            print(f'num generate: {num_generate}')
            for rare_type in rare_types:
                modifier = f'include \'{rare_type}\''
                generated_data.extend(self._generate_modified(data, modifier, num_generate))
        if return_metrics:
            return generated_data, metrics
        return generated_data


from powerml import Generator
from powerml import WriteEmailModel

class EmailGenerator(Generator):
    '''
    This is a class that can be used to generate more data for WriteEmailModels.
    '''

    def __init__(self, gold_types=['Announcement']):
        super().__init__(gold_types)

    def coverage_generator(self, data, return_metrics=True):
        return super().coverage_generator(data, return_metrics, WriteEmailModel())


from datetime import datetime
import time
from powerml import PowerML
from powerml.utils.run_ai import run_ai
from powerml.utils.generator import get_types
from powerml.utils.constants import PARALLEL_REQUEST_LIMIT, DELAY_IN_SECONDS
from random import sample
from math import ceil

class Generator():
    '''
    This is a general class that can be used to generate examples that are not already covered by data.
    '''

    def __init__(self, gold_types=['Llama'], model='text-davinci-002', max_output_tokens=256):
        self.model = model
        self.max_output_tokens = max_output_tokens
        assert len(gold_types) >= 1, 'Please provide at least 1 gold type.'
        self.gold_types = gold_types

    def __fuzzy_modify(self, data, prompt): # NOTE: could add reference data for this
        modified_examples = []
        for i in range(0, len(data), PARALLEL_REQUEST_LIMIT):
            curr_data = data[i:i+PARALLEL_REQUEST_LIMIT]
            prompt_append = f'\n\nModify this example to {prompt}.\n\n\"'
            prompts = [f'\"{datum}\"{prompt_append}' for datum in curr_data]
            generations = run_ai(prompts,
                            stop='\"',
                            api="openai",
                            model=self.model,
                            max_tokens=self.max_output_tokens,
                            temperature=0.5,
                            )
            if type(generations) != list:
                generations = [generations]
            modified_examples.extend([generation.strip() for generation in generations])
            time.sleep(DELAY_IN_SECONDS)
        return modified_examples

    def _generate_modified(self, data, modifier, num_generate):
        print('Start Generating Modified Data:', datetime.now())
        generated_data = self.__fuzzy_modify(sample(data, num_generate), modifier)
        print('End Generating Modified Data:', datetime.now())
        return generated_data

    def compute_coverage(self, generated_types):
        generated_types = set(generated_types)
        gold_types = set(self.gold_types)
        num_matched_items = len(gold_types.intersection(generated_types))
        num_real_items = len(gold_types)
        coverage = num_matched_items / num_real_items
        rare_types = gold_types.difference(generated_types)
        return {
            'Coverage': coverage,
            'Rare Types': rare_types,
        }

    def coverage_generator(self, data, return_metrics=True, prediction_model=PowerML()):
        model = prediction_model
        # should call model.fit(data) once available
        model_predictions = []
        for datum in data:
            model_predictions.append(model.predict(datum))
        print(f'model predictions: {model_predictions}')
        generated_types = get_types(model_predictions, self.gold_types)
        metrics = self.compute_coverage(generated_types)
        coverage = metrics['Coverage']
        rare_types = metrics['Rare Types']
        generated_data = []
        if rare_types:
            # generate at least 1 example per rare_type, proportional to the amount of coverage of the data
            num_generate = ceil((1 - coverage) * len(data) / len(rare_types))
            print(f'num generate: {num_generate}')
            for rare_type in rare_types:
                modifier = f'include \'{rare_type}\''
                generated_data.extend(self._generate_modified(data, modifier, num_generate))
        if return_metrics:
            return generated_data, metrics
        return generated_data



from powerml import Filter

class QuestionFilter(Filter):
    '''
    This is a class that can be used to filter noise from data for QuestionAnswerModels.
    '''

    def __init__(self):
        super().__init__('text-davinci-002', 'question for students of the lesson')



from powerml import PowerML
import logging
import re
logger = logging.getLogger(__name__)


class AutocompleteSQLModel:
    def __init__(self,
                 config={},
                 max_output_tokens=256,
                 ):
        self.model = PowerML(config)
        self.model_name = "hex/sql-autocomplete/v2"
        self.max_output_tokens = max_output_tokens
        self.table_schemas = []
        self.example_queries = []

    def fit(self, table_schemas=[], example_queries=[]):
        """
        Parameters
        ----------
        table_schemas : list
            Takes in a list of schema definitions where element of the
            list corresponds to the schema for one table.
        example_queries: list
            Takes a list of sql queries that have been used with the tables
            to get sql results.
        """
        self.table_schemas = table_schemas
        self.example_queries = example_queries

    def predict(self, input):
        prompt = {
            "{{input}}": input,
            "{{examples}}": "\nEND".join(self.example_queries),
            "{{table_schemas}}": "\n".join(self.table_schemas)
        }

        output = self.model.predict(
            prompt,
            max_tokens=self.max_output_tokens,
            stop=['\\\\nEND', '\\nEND', '\nEND', ';'],
            temperature=0.7,
            model=self.model_name)
        print(output)
        return self.post_process(output)

    def post_process(self, output):
        # TODO: replace with stop tokens
        results = re.split('\\\\nEND|\\nEND|\nEND|select|;', output)
        return results[0].strip()



'''Constants'''

MAX_SHORT_TOKENS = 20
N_LOGPROBS = 5
LOGPROB_THRESHOLD = -7
PARALLEL_REQUEST_LIMIT = 1 # ideally would like to set this to 20, but would require changes to run_ai
PER_MINUTE_REQUEST_LIMIT = 20
DELAY_IN_SECONDS = 60.0 / PER_MINUTE_REQUEST_LIMIT
KEYWORDS = ['true', 'null', '*', 'select', 'from', 'where', 'with', 'group by', 'inner', 'outer', 'left', 'join']


from powerml.utils.constants import KEYWORDS, DELAY_IN_SECONDS
from powerml.fuzzy.fuzzy_contains import fuzzy_contains
import json
from collections import defaultdict
from sql_metadata import Parser
import time

def get_types(model_predictions, gold_types):
    generated_types = set()
    for model_prediction in model_predictions:
        for gold_type in gold_types:
            if gold_type not in generated_types:
                is_contained = fuzzy_contains(model_prediction, gold_type)
                if is_contained:
                    time.sleep(DELAY_IN_SECONDS)
                    generated_types.add(gold_type)
    return generated_types

def load_menu(menu_filename):
    items = []
    with open(menu_filename) as menu_file:
        menu_data = json.load(menu_file)
        for menu_datum in menu_data:
            items.append(menu_datum['name'])
    return items

def load_menu_with_associated_aliases(menu_filename, alias_filename):
    with open(alias_filename) as alias_file:
        alias = json.load(alias_file)
        menu = load_menu(menu_filename)
        menu_with_aliases = {}
        for item in menu:
            menu_with_aliases[item] = item
            if item in alias:
                if 'display_name' in alias[item]:
                    menu_with_aliases[alias[item]['display_name']] = item
                if 'extra_display_name' in alias[item]:
                    menu_with_aliases[alias[item]['extra_display_name']] = item
    return menu_with_aliases

def load_menu_from_files():
    menu_filename = 'deltaco.json'
    alias_filename = 'deltaco_props.json'
    return load_menu_with_associated_aliases(
        menu_filename, alias_filename)

def create_schema(queries):
        tables = defaultdict(set)
        for query in queries:
            try:
                query_parser = Parser(query)
                if query.startswith('select'):
                    table = query_parser.tables
                    if not table:
                        from_index = query.find(' from ') + len(' from ')
                        where_index = query.find(' where ')
                        table = query[from_index:where_index]
                    else:
                        table = table[0]
                    tables[table].update({column for column in query_parser.columns if column not in KEYWORDS})
                else: # query.startswith('with')
                    for with_query in query_parser.with_queries.values():
                        with_query_parser = Parser(with_query)
                        table = with_query_parser.tables[0]
                        if table not in query_parser.with_names:
                            tables[table].update({column.split('.')[-1] for column in with_query_parser.columns if column not in KEYWORDS})
            except:
                continue
        return tables

def tables_to_schema(tables):
    schema = []
    for table in tables:
        columns = ',\n    '.join(tables[table]) + ','
        schema.append(f"""create table {table} (\n    {columns}\n);""")
    return schema


from datetime import datetime
import time
from powerml.utils.run_ai import run_ai
from powerml.utils.constants import PARALLEL_REQUEST_LIMIT, DELAY_IN_SECONDS

class Filter():
    '''
    This is a general class that can be used to filter noise from data.
    '''

    # NOTE: Would like to add a general filter for near duplicates using Greg's Presto method
    general_filters = {}

    def __init__(self, model, prompt):
        self.model = model
        self.prompt = prompt
        self.filters = dict(self.general_filters, Invalid=self.__filter_invalid)

    def __add_input(self, data):
        if type(data[0]) in [tuple, list]:
            return [[datum_input + '\n', datum_output] for datum_input, datum_output in data]
        return [['', datum] for datum in data]

    def __fuzzy_is_valid(self, data): # NOTE: could add reference data for this
        valid_examples = []
        for i in range(0, len(data), PARALLEL_REQUEST_LIMIT):
            curr_data = data[i:i+PARALLEL_REQUEST_LIMIT]
            prompt_prepend = f'Is this a valid {self.prompt}:\n\n'
            prompt_append = '\n\nAnswer Yes if it is valid and No if it is not valid, and explain why or why not.'
            prompts = [f'{prompt_prepend}{datum_input}\"{datum_output}\"{prompt_append}' for datum_input, datum_output in self.__add_input(curr_data)]
            is_valid = run_ai(prompts,
                            stop='\nEND',
                            api="openai",
                            model=self.model,
                            max_tokens=6,
                            temperature=0.0,
                            )
            if type(is_valid) != list:
                is_valid = [is_valid]
            is_valid = [False if 'no' in output.strip().lower() else True for output in is_valid]
            valid_examples.extend([datum for j, datum in enumerate(curr_data) if is_valid[j]])
            time.sleep(DELAY_IN_SECONDS)
        return valid_examples

    def __filter_invalid(self, data):
        print('Start Filtering Invalid Data:', datetime.now())
        filtered_data = self.__fuzzy_is_valid(data)
        print('End Filtering Invalid Data:', datetime.now())
        return filtered_data

    def noise_filter(self, data, return_metrics=True):
        metrics = {}
        len_data = len(data)
        filtered_data_intersect = list(dict.fromkeys(data).keys())
        metrics[f'Duplicate Noise'] = 1 - len(filtered_data_intersect) / len_data
        for filter in self.filters:
            filtered_data = self.filters[filter](data)
            metrics[f'{filter} Noise'] = 1 - len(filtered_data) / len_data
            filtered_data_intersect = [datum for datum in filtered_data_intersect if datum in filtered_data]
        metrics['Total Noise'] = 1 - len(filtered_data_intersect) / len_data
        if return_metrics:
            return filtered_data_intersect, metrics
        return filtered_data_intersect


from powerml import PowerML
import re
from collections import Counter
import random
from tqdm import tqdm
import sys
import os


class FilterTopicsModel:
    '''
    This model removes topics that are not in the
    topic_type category.
    '''

    def __init__(
            self,
            config={},):
        self.model = PowerML(config)
        self.filter_model_name = "unblocked/filter-topics"
        self.memo_filtered_topics = {}
        self.topics = []

    def fit(self, topics, topic_type):
        self.topic_type = topic_type
        self.topics = topics

    def predict(self):
        # If get_topics has been called on the same messages array
        # Then use the previous results
        hash_docs = hash(frozenset(self.topics))
        if hash_docs in self.memo_filtered_topics:
            return self.memo_filtered_topics[hash_docs]

        not_topics = self.__filter_top_topics(self.topics)
        filtered_top_topics = list(set(self.topics) - set(not_topics))

        self.memo_filtered_topics[hash_docs] = filtered_top_topics
        return filtered_top_topics

    def __filter_top_topics(self, topics_in_list):
        prompt = {
            "{{topics_in_list}}": f"{topics_in_list}", "{{topic_type}}": self.topic_type}
        output = self.model.predict(
            prompt, max_tokens=500, temperature=0.7,  model=self.filter_model_name)
        output = self.__parse_output(output)
        return output

    def __parse_output(self, output):
        list_pattern = re.compile(r"\d+\.\s")
        # include enumerated list prompt
        items = list_pattern.sub("", f'1. {output}')
        parsed = []
        for i in items.split('\n'):
            ii = i.split(',')
            stripped = [iii.strip().replace('.', '') for iii in ii if iii]
            parsed.extend(stripped)
        return parsed



from powerml.model.PowerML import PowerML
from powerml.model.ForecastSequenceModel import ForecastSequenceModel
from powerml.model.AutocompleteSQLModel import AutocompleteSQLModel
from powerml.model.WriteEmailModel import WriteEmailModel
from powerml.model.ExtractMenuItemsModel import ExtractMenuItemsModel
from powerml.model.CreateTopicsModel import CreateTopicsModel
from powerml.model.ExtractTopicsModel import ExtractTopicsModel
from powerml.model.FilterTopicsModel import FilterTopicsModel
from powerml.model.QuestionAnswerModel import QuestionAnswerModel
from powerml.filters.Filter import Filter
from powerml.filters.ForecastFilter import ForecastFilter
from powerml.filters.SQLFilter import SQLFilter
from powerml.filters.EmailFilter import EmailFilter
from powerml.filters.MenuFilter import MenuFilter
from powerml.filters.TopicFilter import TopicFilter
from powerml.filters.QuestionFilter import QuestionFilter
from powerml.generators.Generator import Generator
from powerml.generators.TopicGenerator import TopicGenerator
from powerml.generators.MenuGenerator import MenuGenerator
from powerml.generators.SQLGenerator import SQLGenerator
from powerml.generators.EmailGenerator import EmailGenerator
from powerml.generators.ForecastGenerator import ForecastGenerator
from powerml.generators.QuestionGenerator import QuestionGenerator


from powerml import Generator
from powerml import ExtractMenuItemsModel
from powerml.utils.generator import load_menu_from_files
from math import ceil
import re

class MenuGenerator(Generator):
    '''
    This is a class that can be used to generate more data for ExtractMenuItemsModels.
    '''

    def __init__(self, menu=None):
        self.menu = menu
        if not self.menu:
            self.menu = load_menu_from_files()
        super().__init__(list(self.menu.values()))

    def __strip_count(self, name_and_count):
        match = re.search(r"^\d+x", name_and_count)

        if match is None:
            return name_and_count

        return name_and_count[match.end(0):].strip()
    
    def __convert_item(self, item):
        item = self.__strip_count(item)
        return self.menu[item] if item in self.menu else item

    def __order_to_items(self, order):
        items = [self.__convert_item(item.lstrip()) for item in order.split('\n')]
        return items

    def coverage_generator(self, data, return_metrics=True):
        model = ExtractMenuItemsModel()
        # should call model.fit(data) once available
        generated_items = set()
        for datum in data:
            generated_items.update(self.__order_to_items(model.predict(datum)))
        metrics = self.compute_coverage(generated_items)
        coverage = metrics['Coverage']
        rare_items = metrics['Rare Types']
        generated_data = []
        if rare_items:
            # generate at least 1 example per rare_item, proportional to the amount of coverage of the data
            num_generate = ceil((1 - coverage) * len(data) / len(rare_items))
            print(f'num generate: {num_generate}')
            for item in rare_items:
                modifier = f'include 1x \'{item}\''
                generated_data.extend(self._generate_modified(data, modifier, num_generate))
        if return_metrics:
            return generated_data, metrics
        return generated_data